@@ -29,12 +29,15 @@ import os
 import json
 import logging
 from functools import wraps
-from typing import List, Dict, Optional, Any, Literal, Union, Set
+from typing import List, Dict, Optional, Any, Literal, Union, Set, Type, TypedDict
 
 from oauthlib.oauth2 import BackendApplicationClient
 from requests_oauthlib import OAuth2Session
 
 from . import status
+from .api_actions import Actions
+from .api_functions import Functions
+from .utils import sentinel
 from .exceptions import (
     DynamicsException,
     ParseError,
@@ -49,10 +52,13 @@ from .exceptions import (
     WebAPIUnavailable,
 )
 
+try:
+    from django.core.cache import cache
+except ImportError:
+    from .utils import cache
 
-__all__ = [
-    "DynamicsClient",
-]
+
+__all__ = ["DynamicsClient"]
 
 
 logger = logging.getLogger(__name__)
@@ -61,9 +67,18 @@ method_type = Literal["get", "post", "patch", "delete"]
 orderby_type = Dict[str, Literal["asc", "desc"]]
 filter_type = Union[Set[str], List[str]]
 
+
+class ExpandType(TypedDict):
+    select: List[str]
+    filter: filter_type
+    top: int
+    orderby: orderby_type
+    expand: Dict[str, "ExpandType"]
+
+
 expand_keys = Literal["select", "filter", "top", "orderby", "expand"]
-expand_values = Union[List[str], Set[str], int, orderby_type, Dict]
-expand_type = Dict[str, Dict[expand_keys, expand_values]]
+expand_values = Union[List[str], Set[str], int, orderby_type, Dict[str, ExpandType]]
+expand_type = Dict[str, ExpandType]
 
 
 def error_simplification_available(func):
@@ -80,13 +95,13 @@ def error_simplification_available(func):
     @wraps(func)
     def inner(*args, **kwargs):
         simplify_errors: bool = kwargs.pop("simplify_errors", False)
-        raise_separately: List[Exception] = kwargs.pop("raise_separately", [])
+        raise_separately: List[Type[Exception]] = kwargs.pop("raise_separately", [])
 
         try:
             return func(*args, **kwargs)
         except Exception as error:
             logger.error(error)
-            if not simplify_errors or type(error) in raise_separately:
+            if not simplify_errors or any([isinstance(error, exception) for exception in raise_separately]):
                 raise error
             else:
                 raise DynamicsException("There was a problem communicating with the server.")
@@ -94,13 +109,6 @@ def error_simplification_available(func):
     return inner
 
 
-class _sentinel:
-    """Sentinel value."""
-
-    def __bool__(self):
-        return False
-
-
 class DynamicsClient:
     """Dynamics client for making queries from a Microsoft Dynamics 365 database."""
 
@@ -132,16 +140,23 @@ class DynamicsClient:
                       Most likely in this format: https://[Organization URI]/.default
         """
 
-        # [sic] Assure that url ends in forward slash
         self._api_url = api_url.rstrip("/") + "/"
+        self._session = OAuth2Session(client=BackendApplicationClient(client_id=client_id))
+        self._session.fetch_token(token_url=token_url, client_id=client_id, client_secret=client_secret, scope=scope)
 
-        # [sic] Scope not given here, since we don't have a token yet
-        self._api = OAuth2Session(client=BackendApplicationClient(client_id=client_id))
+        self._session = OAuth2Session(client=BackendApplicationClient(client_id=client_id))
+        token = cache.get("dynamics-client-token", None)
+        if token is None:
+            token = self._session.fetch_token(token_url=token_url, client_secret=client_secret, scope=scope)
+            cache.set("dynamics-client-token", token, int(token.get("expires_in", 3599)) - 60)
+        else:
+            self._session.token = token
 
-        # Fetch token
-        self._api.fetch_token(token_url=token_url, client_id=client_id, client_secret=client_secret, scope=scope)
+        self.actions = Actions(client=self)
+        """Predefined dynamics API actions. Used to make certain changes to data."""
+        self.functions = Functions(client=self)
+        """Predefined dynamics API functions. Used to fetch certain types of data."""
 
-        # Query options
         self._select: List[str] = []
         self._expand: expand_type = {}
         self._filter: filter_type = []
@@ -154,9 +169,10 @@ class DynamicsClient:
         self._row_id = ""
         self._add_ref_to_property = ""
         self._pre_expand = ""
+        self._apply = ""
 
         self._headers: Dict[str, str] = {}
-        self._pagesize: int = 1000
+        self._pagesize: int = 5000
 
     def __getitem__(self, key):
         return self.headers[key]
@@ -168,7 +184,7 @@ class DynamicsClient:
     def from_environment(cls):
         """Create a client from environment variables:
 
-        * DYNAMICS_BASE_URL (api_url): url string
+        * DYNAMICS_API_URL: url string
         * DYNAMICS_TOKEN_URL: url string
         * DYNAMICS_CLIENT_ID: client id string
         * DYNAMICS_CLIENT_SECRET: client secret key string
@@ -177,7 +193,7 @@ class DynamicsClient:
         :raises KeyError: An environment variable was not configured properly
         """
 
-        base_url = os.environ["DYNAMICS_BASE_URL"]
+        base_url = os.environ["DYNAMICS_API_URL"]
         token_url = os.environ["DYNAMICS_TOKEN_URL"]
         client_id = os.environ["DYNAMICS_CLIENT_ID"]
         client_secret = os.environ["DYNAMICS_CLIENT_SECRET"]
@@ -199,7 +215,7 @@ class DynamicsClient:
             if query[-1] != "/":
                 query += "/"
             query += self.action
-        if self.add_ref_to_property and not self.pre_expand and not self.action:
+        if self.add_ref_to_property and not any([self.pre_expand, self.action]):
             query += f"/{self.add_ref_to_property}/$ref"
         if (qo := self._compile_query_options()) and not self.add_ref_to_property:
             query += qo
@@ -212,6 +228,7 @@ class DynamicsClient:
                 statement
                 for statement in [
                     self._compile_expand(),
+                    self._compile_apply(),
                     self._compile_select(),
                     self._compile_filter(),
                     self._compile_top(),
@@ -243,6 +260,7 @@ class DynamicsClient:
         self._row_id = ""
         self._add_ref_to_property = ""
         self._pre_expand = ""
+        self._apply = ""
 
         self._headers: Dict[str, str] = {}
 
@@ -285,7 +303,7 @@ class DynamicsClient:
 
         Please also read the decorator's documentation!
 
-        :param not_found_ok: No entities found should not raise NotFound error, but return empty list instead.
+        :param not_found_ok: No entities returned should not raise NotFound error, but return empty list instead.
         :param next_link: Request the next set of records from this link instead.
         """
 
@@ -293,9 +311,9 @@ class DynamicsClient:
         self.set_default_headers("get")
 
         if next_link is not None:
-            response = self._api.get(next_link, headers=self.headers)
+            response = self._session.get(next_link, headers=self.headers)
         else:
-            response = self._api.get(self.current_query, headers=self.headers)
+            response = self._session.get(self.current_query, headers=self.headers)
 
         data = response.json()
 
@@ -318,8 +336,8 @@ class DynamicsClient:
                 if "@odata.nextLink" in column_key:
 
                     # Sometimes @odata.next links will appear even if all items were fetched.
-                    # We set odata.maxpagesize to 1000 so the first 1000 items should be fetched.
-                    # Therefore, if the @odata.next link appears before that, we can ignore it.
+                    # We know how many items should be fetched from odata.maxpagesize header,
+                    # therefore, if the @odata.next link appears before that, we can ignore it.
                     #
                     key = column_key[:-15]
                     if len(entities[i][key]) < self.pagesize:
@@ -330,7 +348,7 @@ class DynamicsClient:
                     # page of data as well, so we filter those out. Although, This doesn't seem
                     # to be the intended way this should work, based on this:
                     #
-                    # https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/query-data-web-api#specify-the-number-of-tables-to-return-in-a-page
+                    # https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/query-data-web-api#retrieve-related-tables-by-expanding-navigation-properties
                     #
                     extra = self.GET(not_found_ok=not_found_ok, next_link=row.pop(column_key))
                     id_tags = [value["@odata.etag"] for value in entities[i][key]]
@@ -345,8 +363,7 @@ class DynamicsClient:
 
     @error_simplification_available
     def POST(self, data: Dict[str, Any]) -> Dict[str, Any]:
-        """Create new row in a table or execute an API action or function.
-        Must have 'table' query option set.
+        """Create new row in a table. Must have 'table' attribute set. Use expand and select to reduce returned data.
 
         Please also read the decorator's documentation!
 
@@ -356,9 +373,8 @@ class DynamicsClient:
         self.request_counter += 1
         self.set_default_headers("post")
 
-        # [sic] POSTing data in dict form doesn't work for some reason...
         data = json.dumps(data).encode()
-        response = self._api.post(self.current_query, data=data, headers=self.headers)
+        response = self._session.post(self.current_query, data=data, headers=self.headers)
 
         if response.status_code == status.HTTP_204_NO_CONTENT:
             return {}
@@ -372,7 +388,7 @@ class DynamicsClient:
 
     @error_simplification_available
     def PATCH(self, data: Dict[str, Any]) -> Dict[str, Any]:
-        """Update row in a table. Must have 'table' and 'row_id' query option set.
+        """Update row in a table. Must have 'table' and 'row_id' attributes set.
         Use expand and select to reduce returned data.
 
         Please also read the decorator's documentation!
@@ -384,7 +400,7 @@ class DynamicsClient:
         self.set_default_headers("patch")
 
         data = json.dumps(data).encode()
-        response = self._api.patch(self.current_query, data=data, headers=self.headers)
+        response = self._session.patch(self.current_query, data=data, headers=self.headers)
 
         if response.status_code == status.HTTP_204_NO_CONTENT:
             return {}
@@ -398,7 +414,7 @@ class DynamicsClient:
 
     @error_simplification_available
     def DELETE(self) -> Dict[str, Any]:
-        """Delete row in a table. Must have 'table' and 'row_id' query option set.
+        """Delete row in a table. Must have 'table' and 'row_id' attributes set.
 
         Please also read the decorator's documentation!
         """
@@ -406,7 +422,7 @@ class DynamicsClient:
         self.request_counter += 1
         self.set_default_headers("delete")
 
-        response = self._api.delete(self.current_query, headers=self.headers)
+        response = self._session.delete(self.current_query, headers=self.headers)
 
         if response.status_code == status.HTTP_204_NO_CONTENT:
             return {}
@@ -443,7 +459,8 @@ class DynamicsClient:
         ...and how to Use Web API Actions:
         https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/use-web-api-actions.
 
-        You can input the action/function as a string, or use the included `act` and `fnc` objects construct it.
+        Most of the time you don't need to set this, since you can use the .actions and .functions attributes
+        to make these queries.
         """
         return self._action
 
@@ -523,8 +540,8 @@ class DynamicsClient:
         """Set $select statement. Select which columns are returned from the table."""
         self._select = items
 
-    def _compile_select(self, values: List[str] = _sentinel) -> str:
-        if values is _sentinel:
+    def _compile_select(self, values: List[str] = sentinel) -> str:
+        if values is sentinel:
             values = self._select
 
         return "$select=" + ",".join([key for key in values]) if values else ""
@@ -558,8 +575,8 @@ class DynamicsClient:
 
         self._expand = items
 
-    def _compile_expand(self, items: expand_type = _sentinel) -> str:
-        if items is _sentinel:
+    def _compile_expand(self, items: expand_type = sentinel) -> str:
+        if items is sentinel:
             items = self._expand
 
         if not items:
@@ -605,7 +622,7 @@ class DynamicsClient:
         ...and how to Query data using the Web API:
         https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/query-data-web-api.
 
-        You can input the filters as strings, or use the included `ftr` object to construct them.
+        You can input the filters as strings, or use the included `ftr`-object to construct them.
 
         Below is a list of the standard operators:
 
@@ -633,17 +650,42 @@ class DynamicsClient:
 
         self._filter = items
 
-    def _compile_filter(self, values: filter_type = _sentinel):
-        if values is _sentinel:
+    def _compile_filter(self, values: filter_type = sentinel):
+        if values is sentinel:
             values = self._filter
 
         if not values:
             return ""
 
         if isinstance(values, set):
-            return "$filter=" + " or ".join([key.strip() for key in values])
+            return "$filter=" + " or ".join([value.strip() for value in values])
         elif isinstance(values, list):
-            return "$filter=" + " and ".join([key.strip() for key in values])
+            return "$filter=" + " and ".join([value.strip() for value in values])
+
+    @property
+    def apply(self):
+        """Current apply statement."""
+        return self._apply
+
+    @apply.setter
+    def apply(self, statement: str):
+        """Set the $apply statement. Aggregates or groups results.
+
+        It is recommended to read how to aggregate and grouping results:
+        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/webapi/query-data-web-api#aggregate-and-grouping-results
+
+        ...and the FetchXML aggregation documentation:
+        https://docs.microsoft.com/en-us/powerapps/developer/data-platform/use-fetchxml-aggregation
+
+        You can input the apply-statement as a string, or use the included `apl`-object to construct it.
+
+        :param statement: aggregate, groupby, or filter apply-string.
+        """
+
+        self._apply = statement
+
+    def _compile_apply(self):
+        return f"$apply={self._apply}"
 
     @property
     def top(self) -> int:
@@ -657,8 +699,8 @@ class DynamicsClient:
         """
         self._top = number
 
-    def _compile_top(self, number: int = _sentinel) -> str:
-        if number is _sentinel:
+    def _compile_top(self, number: int = sentinel) -> str:
+        if number is sentinel:
             number = self._top
 
         return f"$top={number}" if number != 0 else ""
@@ -679,8 +721,8 @@ class DynamicsClient:
 
         self._orderby = items
 
-    def _compile_orderby(self, values: orderby_type = _sentinel) -> str:
-        if values is _sentinel:
+    def _compile_orderby(self, values: orderby_type = sentinel) -> str:
+        if values is sentinel:
             values = self._orderby
 
         if not values:
@@ -701,8 +743,8 @@ class DynamicsClient:
         """
         self._count = value
 
-    def _compile_count(self, value: bool = _sentinel) -> str:
-        if value is _sentinel:
+    def _compile_count(self, value: bool = sentinel) -> str:
+        if value is sentinel:
             value = self._count
 
         return f"$count=true" if value else ""
@@ -714,7 +756,7 @@ class DynamicsClient:
 
     @pagesize.setter
     def pagesize(self, value: int):
-        """Specify the number of tables to return in a page. By default, this is set to 1000. Maximum is 5000."""
+        """Specify the number of tables to return in a page."""
 
         if value < 1:
             ValueError(f"Value must be bigger than 0. Got {value}.")
@@ -722,6 +764,3 @@ class DynamicsClient:
             ValueError(f"Max pagesize is 5000. Got {value}.")
 
         self._pagesize = value
-
-    # TODO: apply-statement
-    # TODO: Any and all statements
